#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <ctype.h>

#ifndef PI
  #define PI 3.14159265358979323
#endif
#ifndef FALSE
  #define FALSE 0
#endif
#ifndef TRUE
  #define TRUE 1
#endif


/*
#define RealSinc(d) ((fabs(d)<0.0000001)?1.0:sin(d)/d)
*/

static __inline double RealSinc(double d) {
    if (fabs(d) < 0.0000001)
	return 1.0;
    else
	return sin(d)/d;
}

static double sinc(double d) {
    static double min=1.0, max=-1.0;
    static double dist = 0.0;
    static double *table = NULL;
    const int itemsPerUnit = 128;
    static long items = 0;
    if (d < min || d > max) {
	int i;
	if (d < min)
	    min = d - 10;
	if (d > max)
	    max = d + 10;
	if (table)
	    free(table);
	dist = max-min;
	items = (long) (dist*itemsPerUnit);
	table = malloc(items*sizeof(double));
	if (!table)
	    exit(EXIT_FAILURE);
	for(i = 0; i < items; i++) {
	    table[i] = RealSinc(min + (double)i/itemsPerUnit);
	}
    }
    return table[(long)((d-min)*itemsPerUnit)];
}



static void ScaleVector(double *in, double *out, int inSize, int outSize,
			int dist) {
    int i, j;
    double multiplier;
    int start, end;

    if (inSize < 1 || outSize < 1)
	return;
    if (dist < 2)
	dist = 2;

    multiplier = (double)(inSize)/(double)(outSize);

    for (j = 0; j < outSize; j++) {
	out[j] = 0.0;
	start = (int) (j*multiplier + 0.5) - dist;
	end   = (int) (j*multiplier + 0.5) + dist;
/*	fprintf(stderr, "Start = %d, End = %d, out = %d, in = %d, j = %d\n",
		start, end, outSize, inSize, j);*/
	i = start;
	while(i < 0) {
	    double s = sinc((j*multiplier-i)*PI);
	    out[j] += in[0]*s;
	    i++;
	}
	while (i < inSize && i <= end) {
	    double s = sinc((j*multiplier-i)*PI);
	    out[j] += in[i++]*s;
	}
	while (i <= end) {
	    double s = sinc((j*multiplier-i)*PI);
	    out[j] += in[inSize-1]*s;
	    i++;
	}
    }
}

#if 0
static void ScaleVectoIdeal(double *in, double *out, int inSize, int outSize) {
    int i, j;
    double multiplier;

    if (inSize < 1 || outSize < 1)
	return;

    multiplier = (double)(inSize)/(double)(outSize);

    for (j = 0; j < outSize; j++) {
	out[j] = 0.0;
	for (i = 0; i < inSize; i++) {
	    double s = sinc((j*multiplier-i)*PI);
	    out[j] += in[i]*s;
	}
    }
}
#endif

static long ReadUnsignedInt(FILE *in) {
    long val = 0;
    int ch = fgetc(in);

    while (isspace(ch))
	ch = fgetc(in);
    while (isdigit(ch)) {
        val = val*10 + ch-'0';
        ch = fgetc(in);
    }
    if (!isspace(ch)) {
	return -1;
    }
    return val;
}


enum PicType {
    PIC_PBM,
    PIC_PGM,
    PIC_PPM,
    PIC_INVALID
};

struct PICTURE {
    enum PicType type;
    long width, height, maxVal;
    unsigned char *data[3];
};


static void FreePicture(struct PICTURE *p) {
    if (p) {
	int i;
	for (i=0; i<3; i++) {
	    if (p->data[i])
		free(p->data[i]);
	}
    }
    free(p);
}

static struct PICTURE *AllocPicture(long width, long height, long maxVal,
				    enum PicType typ) {
    static struct PICTURE *p;
    int i;

    if (width < 1 || height < 1 || maxVal < 1 || maxVal > 255) {
	fprintf(stderr, "Can't allocate picture with bad values.\n"
		"width (must be >= 1) is %ld, height (>= 1) is %ld,"
		"maxVal (1..255) is %ld\n", width, height, maxVal);
	return NULL;
    }

    if (!(p = calloc(1, sizeof(struct PICTURE)))) {
	fprintf(stderr, "Not any memory for picture data.\n");
	return NULL;
    }

    p->type = typ;
    p->width = width;
    p->height = height;
    p->maxVal = maxVal;

    for (i=0; p && i<((typ = PIC_PPM) ? 3 : 1); i++) {
	if (!(p->data[i] =
	      malloc(p->width*p->height*sizeof(*p->data)))) {
	    fprintf(stderr, "No memory for picture data.\n");
	    FreePicture(p);
	}
    }

    return p;
}

static struct PICTURE *ReadFile(char *fileName) {
    int ch = 0;  /* This initialization removes an unneeded warning */
    struct PICTURE *pic = NULL;
    int ok = FALSE;
    int picCode;
    enum PicType typ;
    long w, h, maxVal;
    FILE *fp;

    if (fileName) {
	if (!(fp = fopen(fileName, "rb"))) {
	    fprintf(stderr, "Could not open file.\n");
	    goto cleanup;
	}
    } else {
	fp = stdin;
    }

    if (fgetc(fp) != 'P' || (picCode=fgetc(fp)) == EOF) {
        fprintf(stderr, "File is not a PBM, PGM, or PPM file.\n");
	goto cleanup;
    }

    switch(picCode) {
    case '4':
	typ = PIC_PBM;
	break;
    case '5':
	typ = PIC_PGM;
	break;
    case '6':
	typ = PIC_PPM;
	break;
    default:
	fprintf(stderr, "File is not a binary PBM, PGM or PPM file.\n");
	goto cleanup;
    }

    w = ReadUnsignedInt(fp);
    h = ReadUnsignedInt(fp);
    if (typ != PIC_PBM)
	maxVal = ReadUnsignedInt(fp);
    else
	maxVal = 1;

    if (w < 1 || h < 1 || maxVal < 1) {
	fprintf(stderr, "File corrupted\n");
	goto cleanup;
    }

    pic = AllocPicture(w, h, maxVal, typ);
    if (!pic)
	goto cleanup;

    if (pic->type == PIC_PBM) {
	unsigned char *p = pic->data[0];
	long x, y;
	for (y=0; y<pic->height; y++) {
	    for (x=0; x<pic->width; x++) {
		if (!(x&7)) {
		    if ((ch = getc(fp)) == EOF) {
			fprintf(stderr, "Not enough data in file.\n");
			goto cleanup;
		    }
		}
		*p++ = (ch & 128) ? 0 : 1;
		ch <<= 1;
	    }
	}
    } else {
	long size = pic->width*pic->height;
	if (pic->type == PIC_PGM) {
	    if (fread(pic->data[0], 1, size, fp) != size) {
		fprintf(stderr, "Not enough data in file.\n");
		goto cleanup;
	    }
	} else {
	    long loc, color, size = pic->width*pic->height;
	    for (loc = 0; loc < size; loc++) {
		for (color = 0; color < 3; color++) {
		    if ((ch = getc(fp)) == EOF) {
			fprintf(stderr, "Not enough data in file.\n");
			goto cleanup;
		    } else {
			pic->data[color][loc] = ch;
		    }
		}
	    }
	}
    }

    ok = TRUE;

 cleanup:
    if (fileName && fp)
	fclose(fp);
    if (!ok) {
	FreePicture(pic);
	pic = NULL;
    }
    return pic;
}



static int WriteFile(char *fileName, struct PICTURE *pic) {
    int ok = FALSE;
    FILE *fp;
    char magic;

    if (fileName) {
	if (!(fp = fopen(fileName, "wb"))) {
	    goto cleanup;
	}
    } else {
	fp = stdout;
    }

    switch(pic->type) {
    case PIC_PBM:
	magic = '4';
	break;
    case PIC_PGM:
	magic = '5';
	break;
    case PIC_PPM:
	magic = '6';
	break;
    default:
	goto cleanup;
    }

    if (fprintf(fp, "P%c\n%ld %ld\n", magic, pic->width, pic->height) < 0)
	goto cleanup;

    switch(pic->type) {
    case PIC_PBM:
    {
	unsigned char *p = pic->data[0];
	long xx, x, y;
	int ch = 0;
	for (y=0; y<pic->height; y++) {
	    x=0;
	    while (x<pic->width) {
		ch = 0;
		xx = (x+7 >= pic->width) ? pic->width & 7 : 8;
		x += xx;
		while (xx-- > 0) {
		    ch = (ch << 1) + ((*p++) ? 0 : 1);
		}
		ch <<= ((8-(x&7))&7); /* Pad any leftback bits */

		if (putc(ch, fp) == EOF)
		    goto cleanup;
	    }
	}
	ok = TRUE;
	break;
    }
    case PIC_PGM:
    {
	long size = pic->width*pic->height;
	if (fprintf(fp, "%ld\n", pic->maxVal) < 0)
	    goto cleanup;
	if (fwrite(pic->data[0], 1, size, fp) != size)
	    goto cleanup;
	ok = TRUE;
	break;
    }
    case PIC_PPM:
    {
	long loc, color, size = pic->width*pic->height;
	if (fprintf(fp, "%ld\n", pic->maxVal) < 0)
	    goto cleanup;
	for (loc = 0; loc < size; loc++) {
	    for (color = 0; color < 3; color++) {
		if (putc(pic->data[color][loc], fp) == EOF)
		    goto cleanup;
	    }
	}
	ok = TRUE;
	break;
    }
    default:
	break;
    }

 cleanup:
    if (fileName && fp)
	fclose(fp);
    if (!ok)
	fprintf(stderr, "File couldn't be written\n");
    return ok;
}

static void PbmToPgm(struct PICTURE *pic) {
    int i;
    if (pic->type == PIC_PBM) {
	unsigned char *p = pic->data[0];
	pic->type = PIC_PGM;
	pic->maxVal = 255;
	for(i=0; i<pic->width*pic->height; i++) {
	    *p = (*p) ? 255 : 0;
	    p++;
	}
    }
}


struct PICTURE *ScaleX(struct PICTURE *inPic, long newWidth, int dist) {
    struct PICTURE *outPic=NULL;
    double *inVec = NULL, *outVec = NULL;

    if (newWidth == inPic->width)
	return inPic;

    if (!(outPic = AllocPicture(newWidth, inPic->height, inPic->maxVal,
				inPic->type)) ||
	!(inVec = malloc(sizeof(double) * inPic->width)) ||
	!(outVec = malloc(sizeof(double) * newWidth))) {
	fprintf(stderr, "Can't scale picture on the X axis.\n");
    } else {
	long color;
	for (color = 0; color < ((inPic->type == PIC_PPM) ? 3 : 1); color++) {
	    unsigned char *inP = inPic->data[color];
	    unsigned char *outP = outPic->data[color];
	    long y;
	    for (y=0; y<inPic->height; y++) {
		long x;
		for (x=0; x<inPic->width; x++)
		    inVec[x] = (double) (*inP++);
		ScaleVector(inVec, outVec, inPic->width, newWidth, dist);
		for (x=0; x<newWidth; x++) {
		    if (outVec[x] < 0.0)
			outVec[x] = 0.0;
		    if (outVec[x] > (double) outPic->maxVal)
			outVec[x] = (double) outPic->maxVal;

		    *outP++ = (unsigned char) (outVec[x] + 0.5);
		}
	    }
	}
    }

    if (inVec)
	free(inVec);
    if (outVec)
	free(outVec);
    FreePicture(inPic);
    return outPic;
}



struct PICTURE *ScaleY(struct PICTURE *inPic, long newHeight, int dist) {
    struct PICTURE *outPic=NULL;
    double *inVec = NULL, *outVec = NULL;

    if (newHeight == inPic->height)
	return inPic;

    if (!(outPic = AllocPicture(inPic->width, newHeight, inPic->maxVal,
				inPic->type)) ||
	!(inVec = malloc(sizeof(double) * inPic->height)) ||
	!(outVec = malloc(sizeof(double) * newHeight))) {
	fprintf(stderr, "Can't scale picture on the X axis.\n");
    } else {
	long color;
	for (color = 0; color < ((inPic->type == PIC_PPM) ? 3 : 1); color++) {
	    long x;
	    for (x=0; x<inPic->width; x++) {
		unsigned char *inP = inPic->data[color]+x;
		unsigned char *outP = outPic->data[color]+x;
		long y;
		for (y=0; y<inPic->height; y++) {
		    inVec[y] = (double) *inP;
		    inP += inPic->width;
		}
		ScaleVector(inVec, outVec, inPic->height, newHeight, dist);
		for (y=0; y<newHeight; y++) {
		    if (outVec[y] < 0.0)
			outVec[y] = 0.0;
		    if (outVec[y] > (double) outPic->maxVal)
			outVec[y] = (double) outPic->maxVal;

		    *outP = (unsigned char) (outVec[y] + 0.5);
		    outP += outPic->width;
		}
	    }
	}
    }

    if (inVec)
	free(inVec);
    if (outVec)
	free(outVec);
    FreePicture(inPic);
    return outPic;
}



int distTable[10] = {2, 4, 6, 10, 14, 18, 26, 36, 50, 80};

int main(int argc, char **argv) {
    int i;
    struct PICTURE *pic=NULL;
    double scaleX = HUGE_VAL, scaleY = HUGE_VAL,
    newSizeX = HUGE_VAL, newSizeY = HUGE_VAL;
    char *fileName = NULL;
    int err = FALSE, xySize = FALSE;
    int dist = distTable[3];

    for (i=1; i<argc; i++) {
	if (!strcmp(argv[i], "-xscale") && i < argc-1 && !xySize) {
	    sscanf(argv[++i], "%lf", &scaleX);
	} else if (!strcmp(argv[i], "-yscale") && i < argc-1 && !xySize) {
	    sscanf(argv[++i], "%lf", &scaleY);
	} else if (!strcmp(argv[i], "-xsize") && i < argc-1 && !xySize) {
	    sscanf(argv[++i], "%lf", &newSizeX);
	} else if (!strcmp(argv[i], "-ysize") && i < argc-1 && !xySize) {
	    sscanf(argv[++i], "%lf", &newSizeY);
	} else if (!strcmp(argv[i], "-width") && i < argc-1 && !xySize) {
	    sscanf(argv[++i], "%lf", &newSizeX);
	} else if (!strcmp(argv[i], "-height") && i < argc-1 && !xySize) {
	    sscanf(argv[++i], "%lf", &newSizeY);
	} else if (!strcmp(argv[i], "-xysize") && i < argc-2) {
	    sscanf(argv[++i], "%lf", &newSizeX);
	    sscanf(argv[++i], "%lf", &newSizeY);
	    xySize = TRUE;
	} else if (!strcmp(argv[i], "-q") && i < argc-1) {
	    int quality;
	    sscanf(argv[++i], "%d", &quality);
	    if (quality < 0 || quality > 9) {
		fprintf(stderr, "Quality must be"
			" between 0 and 9 (you gave %d)\n", quality);
	    } else {
		dist = distTable[quality];
	    }
	} else {
	    static int no = 0;
	    double dtmp;
	    switch(++no) {
	    case 1:
		sscanf(argv[i], "%lf", &dtmp);
		if (!dtmp) {
		    fprintf(stderr, "Unknown option \"%s\"\n", argv[i]);
		    no--;
		    err = TRUE;
		} else {
		    scaleY = scaleX = dtmp;
		}
		break;
	    case 2:
		fileName = argv[i];
		break;
	    default:
		fprintf(stderr, "Unknown option \"%s\"\n", argv[i]);
		break;
	    }
	}
    }

    if (!err) {
	pic = ReadFile(fileName);
	if (!pic)
	    err = TRUE;
    }

    if (!err) {
	if (xySize) {
	    double xtmp, ytmp;
	    xtmp = newSizeX / pic->width;
	    ytmp = newSizeY / pic->height;
	    if (xtmp < ytmp)
		newSizeY = pic->height * xtmp;
	    else
		newSizeX = pic->width * ytmp;
	} else {
	    switch(((scaleX != HUGE_VAL) ? 8 : 0) +
		   ((scaleY != HUGE_VAL) ? 4 : 0) +
		   ((newSizeX != HUGE_VAL) ? 2 : 0) +
		   ((newSizeY != HUGE_VAL) ? 1 : 0)) {
	    case 0:
		newSizeX = pic->width;
		newSizeY = pic->height;
		break;
	    case 1:
		newSizeX = pic->width * (newSizeY / pic->height);
		break;
	    case 2:
		newSizeY = pic->height * (newSizeX / pic->width);
		break;
	    case 3:
		break;
	    case 4:
		newSizeX = pic->width;
		newSizeY = pic->height * scaleY;
		break;
	    case 5:
		fprintf(stderr,
			"Defining new height and y scale is not allowed.\n");
		err = TRUE;
		break;
	    case 6:
		newSizeY = pic->height * scaleY;
		break;
	    case 7:
		fprintf(stderr,
			"Defining new height and y scale is not allowed.\n");
		err = TRUE;
		break;
	    case 8:
		newSizeX = pic->width * scaleX;
		newSizeY = pic->height;
		break;
	    case 9:
		newSizeX = pic->width * scaleX;
		break;
	    case 10:
		fprintf(stderr,
			"Defining new width and x scale is not allowed.\n");
		err = TRUE;
		break;
	    case 11:
		fprintf(stderr,
			"Defining new width and x scale is not allowed.\n");
		err = TRUE;
		break;
	    case 12:
		newSizeX = pic->width * scaleX;
		newSizeY = pic->height * scaleY;
		break;
	    case 13:
		fprintf(stderr,
			"Defining new height and y scale is not allowed.\n");
		err = TRUE;
		break;
	    case 14:
		fprintf(stderr,
			"Defining new width and x scale is not allowed.\n");
		err = TRUE;
		break;
	    case 15:
		fprintf(stderr,
			"Defining new height and y scale is not allowed, "
			"neither is defining new width and x scale.\n");
		err = TRUE;
		break;
	    }
	}
    }

    if (!err) {
	PbmToPgm(pic);
	if (pic->width*newSizeX*pic->height + pic->height*newSizeY*newSizeX >
	    pic->height*newSizeY*pic->width + pic->width*newSizeX*newSizeY) {
	    if (!(pic = ScaleY(pic, (long)(newSizeY + 0.5), dist)))
		err = TRUE;
	    else if (!(pic = ScaleX(pic, (long)(newSizeX + 0.5), dist)))
		err = TRUE;
	} else {
	    if (!(pic = ScaleX(pic, (long)(newSizeX + 0.5), dist)))
		err = TRUE;
	    else if (!(pic = ScaleY(pic, (long)(newSizeY + 0.5), dist)))
		err = TRUE;
	}
    }

    if (!err)
	WriteFile(NULL, pic);

    FreePicture(pic);

    return (err) ? EXIT_FAILURE : EXIT_SUCCESS;
}
